(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2011-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Term_Order_Impl
imports 
  Term_Order
  Certification_Monads.Check_Monad
  Auxx.Shows_Literal
begin

locale pre_redtriple_impl =
  pre_redtriple S NS NST for S NS NST :: "('f :: showl, 'v :: showl) trs" + 
  fixes checkS :: "('f, 'v) rule \<Rightarrow> showsl check"
    and checkNS :: "('f, 'v) rule \<Rightarrow> showsl check"
    and checkNST :: "('f, 'v) rule \<Rightarrow> showsl check"
  assumes checkSSound: "isOK (checkS rule) \<Longrightarrow> rule \<in> S" 
    and checkNSSound: "isOK (checkNS rule)  \<Longrightarrow> rule \<in> NS"
    and checkNSTSound: "isOK (checkNST rule)  \<Longrightarrow> rule \<in> NST"

locale redtriple_impl = pre_redtriple_impl + redtriple S NS NST 

locale mono_redtriple_impl = redtriple_impl +
  fixes checkMono :: "showsl check"
  assumes checkMonoSound: "isOK (checkMono) \<Longrightarrow> ctxt.closed S"

locale ce_af_redtriple_impl = redtriple_impl + ce_af_redtriple

locale ce_mono_redtriple_impl = mono_ce_redtriple + mono_redtriple_impl

locale mono_ce_af_redtriple_impl = ce_mono_redtriple_impl + ce_af_redtriple_impl

locale mono_ce_af_redtriple_order_impl = redtriple_order + ce_mono_redtriple_impl + ce_af_redtriple_impl

(* valid, check_strict, check_non_strict, check_non_strict_top, af, check_mono, description *)
record ('f, 'v) redtriple =
  valid :: "showsl check"
  s :: "('f, 'v) rule \<Rightarrow> showsl check"
  ns :: "('f, 'v) rule \<Rightarrow> showsl check"
  nst :: "('f, 'v) rule \<Rightarrow> showsl check"
  af :: "'f af"
  mono_af :: "'f af"
  mono :: "('f, 'v) rule list \<Rightarrow> showsl check"
  desc :: showsl
  not_wst :: "('f \<times> nat) list option"
  not_sst :: "('f \<times> nat) list option"
  cpx :: "('f, 'v) complexity_measure \<Rightarrow> complexity_class \<Rightarrow> showsl check"

hide_const (open) valid s ns nst af mono desc cpx not_wst not_sst

record ('f, 'v) root_redtriple =
  valid :: "showsl check"
  s :: "('f, 'v) rule \<Rightarrow> showsl check"
  ns :: "('f, 'v) rule \<Rightarrow> showsl check"
  nst :: "('f, 'v) rule \<Rightarrow> showsl check"
  af :: "'f af"
  aft :: "'f af"
  desc :: showsl

hide_const (open) valid s ns nst af desc aft

definition no_complexity where
  "no_complexity \<equiv> (\<lambda> _ _. False)"

definition no_complexity_check where
  "no_complexity_check \<equiv> (\<lambda> _ _. error (showsl (STR ''complexity analysis unsupported'')) :: showsl check)"

lemma no_complexity_check[simp]: "(\<lambda> cm cc. isOK(no_complexity_check cm cc)) = no_complexity"
  unfolding no_complexity_check_def no_complexity_def by auto

interpretation cpx_term_rel S no_complexity 
  by (unfold_locales) (auto simp: no_complexity_def)

fun not_subterm_rel_info :: "('f,'v)trs \<Rightarrow> ('f \<times> nat)list option \<Rightarrow> bool" where
  "not_subterm_rel_info rel None = True"
| "not_subterm_rel_info rel (Some ws) =  (\<forall> f i. f \<notin> set ws \<longrightarrow> simple_arg_pos rel f i)"

locale generic_redtriple_impl = 
  fixes generate_redtriple :: "'a \<Rightarrow> ('f :: showl, 'v :: showl) redtriple" 
  assumes generate_redtriple:
    "\<lbrakk>isOK (redtriple.valid (generate_redtriple rp));
    isOK (check_allm (redtriple.s (generate_redtriple rp)) s_list);
    isOK (check_allm (redtriple.ns (generate_redtriple rp)) ns_list);
    isOK (check_allm (redtriple.nst (generate_redtriple rp)) nst_list)\<rbrakk> \<Longrightarrow>
    \<exists> S NS NST. 
    cpx_ce_af_redtriple_order S NS NST 
      (redtriple.af (generate_redtriple rp)) 
      (redtriple.mono_af (generate_redtriple rp)) 
      (\<lambda> cm cc. isOK(redtriple.cpx (generate_redtriple rp) cm cc)) \<and> 
    set s_list \<subseteq> S \<and> 
    set ns_list \<subseteq> NS \<and> 
    set nst_list \<subseteq> NST \<and>
    not_subterm_rel_info NS (redtriple.not_wst (generate_redtriple rp)) \<and>
    not_subterm_rel_info S (redtriple.not_sst (generate_redtriple rp)) \<and>
    (isOK (redtriple.mono (generate_redtriple rp) (s_list @ ns_list @ nst_list)) \<longrightarrow>
    mono_ce_af_redtriple_order S NS NST (redtriple.af (generate_redtriple rp)) \<and> ctxt.closed S)"

definition dummy_redtriple :: "('f,'v)redtriple" where
  "dummy_redtriple = undefined \<lparr> redtriple.valid := error undefined \<rparr>" 

typedef ('f :: showl,'v :: showl,'rp) get_redtriple_type = "{ grt :: 'rp \<Rightarrow> ('f,'v)redtriple.
  generic_redtriple_impl grt}" 
  by (intro exI[of _ "\<lambda> _. dummy_redtriple"], standard, 
      unfold_locales, auto simp: dummy_redtriple_def)

setup_lifting type_definition_get_redtriple_type

lift_definition get_redtriple' :: "('f :: showl,'v :: showl,'rp) get_redtriple_type \<Rightarrow> 'rp \<Rightarrow> ('f,'v)redtriple" 
  is "\<lambda> grt rp. grt rp" .

lemma get_redtriple': "generic_redtriple_impl (get_redtriple' grt)" 
  by transfer

locale generic_root_redtriple_impl = 
  fixes generate_root_redtriple :: "'a \<Rightarrow> ('f :: showl,'v :: showl) root_redtriple" 
  assumes generate_root_redtriple:
    "\<lbrakk>isOK (root_redtriple.valid (generate_root_redtriple rp));
    isOK (check_allm (root_redtriple.s (generate_root_redtriple rp)) s_list);
    isOK (check_allm (root_redtriple.ns (generate_root_redtriple rp)) ns_list);
    isOK (check_allm (root_redtriple.nst (generate_root_redtriple rp)) nst_list)\<rbrakk> \<Longrightarrow>
    \<exists> S NS NST. 
    ce_af_root_redtriple_order S NS NST (root_redtriple.af (generate_root_redtriple rp)) (root_redtriple.aft (generate_root_redtriple rp)) \<and> 
    set s_list \<subseteq> S \<and> 
    set ns_list \<subseteq> NS \<and> 
    set nst_list \<subseteq> NST"

(* Precedences for natural numbers *)
definition prc_nat :: "('f \<times> nat \<Rightarrow> nat) \<Rightarrow> ('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool)"
  where "prc_nat pr \<equiv> \<lambda> f g. let pf = pr f; pg = pr g in (pg < pf, pg \<le> pf)"

definition prl_nat :: "('f \<times> nat \<Rightarrow> nat) \<Rightarrow> 'f \<times> nat \<Rightarrow> bool"
  where "prl_nat pr \<equiv> \<lambda> f. pr f = 0"

abbreviation pr_nat where "pr_nat pr \<equiv> (prc_nat pr, prl_nat pr)"

(* Precedences via natural numbers *)

lemma SN_pr_nat: "SN {(f,g). (pr :: 'f \<times> nat \<Rightarrow> nat) g < pr f}"
  by (rule HOL.subst[of _ _ SN, OF _ SN_inv_image[OF SN_nat_gt, of "pr"]],
  auto simp: inv_image_def)

interpretation precedence_nat: precedence "prc_nat pr" "prl_nat pr" 
  for "pr" :: "'f \<times> nat \<Rightarrow> nat" 
  by (unfold_locales, unfold prc_nat_def prl_nat_def Let_def, insert SN_pr_nat[of "pr"], auto)

end

